/*
 * This is a program to handle a multi-user poker game. It uses sockets in
 * the AF_INET family, so it can handle several players even on different
 * machines. The game is 5-card draw poker.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/fcntl.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <unistd.h>
#include <strings.h>

#include "port.h"
#include "pokerd.h"
#include "util.h"
#include "limits.h"

#define	TRUE		1
#define	FALSE		0
#define	ERROR		-1
#define	SCORE_FILE	"/usr/games/lib/POKERSCORES"

#define	START_CASH	200	/* amount new player starts with */
#define	MAX_PLAYERS	5	/* max # concurrent players 	 */
#define DECK_SIZE 52

PLAYER  player[MAX_PLAYERS];	/* player information	 */
int     n_players = 1;		/* computer is playing	 */
static int main_socket = ERROR;	/* socket to accept connections on */


char    ranks[] = "23456789TJQKA";
char    suits[] = "DCHS";
char	*suits_names[] = {"Diamond", "Club", "Heart", "Spade"};


static void computer_bet(PLAYER * player, int n_players, int n_active,
		int n_raises, int pot, int total, char *temp);
static void tie(PLAYER * player, int *winner, int which);
static void computer_draw(PLAYER *player, char *temp);
static void draw(PLAYER * player, int n_players, int *deck, int d, int first);
static void show_cards(PLAYER * player, int n_players, int all, int test);
static void high_score_list(int s);
static void put_cash(char *name, int id, int cash);
static void payoff(PLAYER * player, int n_players, int winner, int pot);
static void tellall(PLAYER * player, int n_players, char *msg);
static void show(PLAYER * player, char *temp, int i, int j);
static int played_before(char *name, int id);
static int get_cash(void);
static void leave(PLAYER *player, int *n_players);
static char *evaluate_char(PLAYER *player, int which);
static void new_players(PLAYER * player, int *n_players, int block);
static void shuffle(int *deck);
static int betting(PLAYER *player, int n_players, int *pot, int first);
static int whowon(PLAYER * player, int n_players, int test);
static void evaluate(PLAYER *player, int which);
static void score_mkempty(void);

int
main(int argc, char **argv)
{
	int     i, c, d;
	int     deck[52];	/* the cards */
	int     pot;		/* size of current pot	 */
	int     winner;		/* which player won this hand	 */
	int     first = 0;	/* first player to bet, draw, etc. */

	extern void srandom(int);	/* starts random number generator */
	extern long time(time_t *);	/* unique parameter to above */
	extern void crash();	/* if broken pipe, release port (I hope) */
	extern int strcmp(const char *, const char *);
	extern int getpid(void);/* more random stuff */

	signal(SIGPIPE, crash);
	signal(SIGHUP, crash);
	signal(SIGXCPU, SIG_IGN);
	srandom((int) time(0) ^ getpid());

	player[0].name = strsave("Computer III");
	player[0].cash = 200;
	player[0].sittingout = FALSE;
	player[0].wantsout = FALSE;

	score_mkempty();

	for (;;) {
		if (n_players < 2) /* wait for someone to play with */
			new_players(player, &n_players, TRUE);

		pot = 0;
		for (i = 0; i < n_players; i++) {
			if (player[i].sittingout == FALSE) {
				player[i].in = TRUE;
				player[i].bet = ANTE;	/* ante */
				player[i].cash -= ANTE;
				pot += ANTE;
			} else {
				player[i].in = FALSE;
				player[i].bet = (-1);	/* keep tellall()quiet*/
			}
			if (n_players == 2)
				player[i].lonehands += 1;
		}

		tellall(player, n_players, "X");	/* clear all screens */
		tellall(player, n_players, "U");	/* clear all screens */

		shuffle(deck);
		d = 0;		/* next card to draw from deck */
		for (c = 0; c < 5; c++)
			for (i = 0; i < n_players; i++)
				if (player[i].in)
					player[i].cards[c] = deck[d++];

		/* show cards to each player ... */
		show_cards(player, n_players, FALSE, FALSE);
		if (++first >= n_players)
			first = 0;

		if ((winner = betting(player, n_players, &pot, first)) < 0) {
			draw(player, n_players, deck, d, first);
			winner = betting(player, n_players, &pot, first);
			if (winner < 0)
				winner = whowon(player, n_players,
				    argc > 1 && strcmp(argv[1], "-test") == 0);
			else if (argc > 1 && strcmp(argv[1], "-test") == 0)
				whowon(player, n_players, TRUE);

		} else if (argc > 1 && strcmp(argv[1], "-test") == 0)
			whowon(player, n_players, TRUE);

		payoff(player, n_players, winner, pot);
		leave(player, &n_players);
		new_players(player, &n_players, FALSE);
	}
	return 0;
}

/*
 * Send everyone a message.
 */
static void 
tellall(PLAYER * player, int n_players, char *msg)
{
	int     i;

	/* don't send any messages to the computer */
	for (i = 1; i < n_players; i++)
		if (player[i].bet != -1) /* no messages to anyone sitting out */
			writeln(player[i].socket, msg);
}

/*
 * Add current pot to winners available cash. Tell everyone who won.
 */

/* player information */
/* number of players */
/* which player won */
/* how much $ was won */
static void 
payoff(PLAYER * player, int n_players, int winner, int pot)
{
	char    temp[80];

	player[winner].cash += pot;
	sprintf(temp, "W%s", player[winner].name);
	tellall(player, n_players, temp);
}

/*
 * Routines to handle players joining game and quitting.
 */

/* array of player info structures	 */
/* number of players (incl comp) 	 */
/* if nobody wants to join, wait on socket until somebody does. */

static void 
new_players(PLAYER * player, int *n_players, int block)
{
	int     readfds;	/* for select() call 	 */
	char    temp[80];	/* to read player's name */
	int     s;		/* temp holder for new socket	 */
	int     rc;		/* # ready selected sockets */
	int     open_sock(void);/* to open main socket */
	struct sockaddr_in from;/* connection acceptor */
	int     fromlen = sizeof(from);
	struct timeval t;	/* don't let select() run forever */
	int     i;

	if (main_socket == ERROR)
		main_socket = open_sock();

	/* set up 1 second timeout for use on all select() calls */
	t.tv_sec = 1L;
	t.tv_usec = 0L;

	/* see if people who were sitting out want to rejoin yet */
	for (i = 1; i < *n_players; i++) {
		if (player[i].sittingout) {
			readfds = 1 << player[i].socket;
			if (select(32, &readfds, 0, 0, &t) > 0)
			{
				readln(player[i].socket, temp);
				if (temp[0] == 'S')
					player[i].sittingout = FALSE;
			}
		}
	}
	for (;;) {
		readfds = 1 << main_socket;
		rc = select(32, &readfds, 0, 0, &t);
		if (rc > 0 || block) {
			if ((s = accept(main_socket, &from, &fromlen)) < 0)
			{
				perror("accept");
				fflush(stdout);
				break;
			}
			if (*n_players >= MAX_PLAYERS) {
				writeln(s, "TOOMANY");
				close(s);
			} else {
				block = FALSE;
				printf("accepted connection from ");
				writeln(s, "OK");

				player[*n_players].socket = s;
				player[*n_players].cash = START_CASH;
				player[*n_players].in = FALSE;
				player[*n_players].wantsout = FALSE;
				player[*n_players].sittingout = FALSE;
				player[*n_players].lonehands = 0;

				readln(s, temp);
				temp[strlen(temp) - 1] = NULL;
				/*
				 * don't allow colons in name because of
				 * format of score file
				 */
				if (index(temp, ':') != NULL)
					*index(temp, ':') = NULL;
				if (strlen(temp) > 25)
					temp[25] = NULL;
				player[*n_players].name = strsave(temp);
				printf("%s\n", temp);
				fflush(stdout);
				readln(s, temp);
				player[*n_players].userid = atoi(temp);
				if (played_before(player[*n_players].name,
				    player[*n_players].userid))
					player[*n_players].cash = get_cash();
				(*n_players)++;
				break;
			}
		} else
			break;
	}
}

static void 
leave(PLAYER * player, int *n_players)
{
	register int i;

	for (i = 1; i < *n_players; i++)
		if (player[i].wantsout) {
			writeln(player[i].socket, "Q");
			put_cash(player[i].name, player[i].userid,
				player[i].cash);
			high_score_list(player[i].socket);
			close(player[i].socket);
			free(player[i].name);
			player[i] = player[*n_players - 1];
			(*n_players)--;
			--i;	/* since this entry is new, test again */
		}
}

/* Open the main socket. */
int 
open_sock(void)
{
	int     s;
	int	flag = 1, ret;
	struct sockaddr_in sockaddr;

	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s < 0) {
		perror("open_sock");
		exit(1);
	}

	ret = setsockopt(s, SOL_SOCKET, SO_REUSEADDR,
		(char *)&flag, sizeof(flag)); /* hunsoo */
	if (ret) {
		perror("setsockopt");
		exit(1);
	}

	sockaddr.sin_family = AF_INET;
	sockaddr.sin_port = htons(PORT);
	sockaddr.sin_addr.s_addr = INADDR_ANY;

	if (bind(s, &sockaddr, sizeof(sockaddr)) < 0) {
		perror("bind");
		exit(1);
	}
	if (listen(s, 0) < 0) {
		perror("listen");
		exit(1);
	}
	return s;
}

void 
crash(void)
{
	int     i;
	for (i = 0; i < 32; i++)
		close(i);
	exit(9);
}


/*
 * Maintain players' cash from one play to the next. Money is stored according
 * to NAME. Thus if a player plays on different systems, he's still playing
 * with the same cash. Also note that if a player changes his name, he starts
 * anew with START_CASH. Userid is also now included, so that someone can not
 * "steal" another player's cash by usurping his name temporarily.
 */


static int scores = (-1);	/* file channel for scores	 */
static char temp[81];

static int
played_before(char *name, int id)
{
	int     found = FALSE;
	char   *p;

#if original
	if (scores == -1)
		if ((scores = open(SCORE_FILE, O_RDWR, 0)) == -1) {
			perror(SCORE_FILE);
			return (FALSE);
		}
#else
	if (scores == -1)
		if ((scores = open(SCORE_FILE, O_RDWR)) == -1) {
			perror(SCORE_FILE);
			return (FALSE);
		}
#endif /* changed by hunsoo */

	lseek(scores, 0, 0);
	while (read(scores, temp, 80) == 80)
		if (strncmp(temp, name, (p = index(temp, ':')) - temp) == 0
		    && atoi(p + 1) == id) {
			found = TRUE;
			break;
		}
	return found;
}

/*
 * Whenever this program restarts, the information of the SCORE_FILE will be
 * gone.--hunsoo@sensor.kaist.ac.kr
 */
static void
score_mkempty(void)
{
	int     fd;

	if (scores != -1) {
		fprintf(stderr, "programmical error!\n");
		exit(1);
	}
	if ((fd = open(SCORE_FILE, O_WRONLY | O_CREAT | O_TRUNC)) == -1) {
		perror(SCORE_FILE);
		exit(1);
	}

	close(fd);
	return;
}


static int
get_cash(void)
{
	return (atoi(index(index(temp, ':') + 1, ':') + 1));
}


/* player's name	 */
/* how much dough he had */
static void 
put_cash(char *name, int userid, int cash)
{
	int     found = FALSE;
	char   *p;
	char    idstr[10];
	lseek(scores, 0, 0);
	while (read(scores, temp, 80) == 80)
		if (strncmp(temp, name, (p = index(temp, ':')) - temp) == 0
		    && atoi(p + 1) == userid) {
			found = TRUE;
			break;
		}

	if (found)
		lseek(scores, -80, 1);
	sprintf(idstr, "%d", userid);
	sprintf(temp, "%s:%s:%-*d\n",
		name, idstr, 77 - strlen(name) - strlen(idstr), cash);
	write(scores, temp, 80);
}

static void 
high_score_list(int s)
{
	struct stuff {
		char    name[30];
		int     score;
		struct stuff *next;
	}      *head, *ptr, *tempptr, *last;
	int     l;
	int     num;
	int     score;
	char    name[30];
	char    output[80];

	head = NULL;
	lseek(scores, 0, 0);
	writeln(s, "HIGH SCORE LIST");
	writeln(s, "=====================================");
	while (read(scores, temp, 80) == 80) {
		strncpy(name, temp, l = (index(temp, ':') - temp));
		name[l] = NULL;
		score = get_cash();
		ptr = head;
		last = NULL;
		while (ptr != NULL && ptr->score > score) {
			last = ptr;
			ptr = ptr->next;
		}
		if (head == NULL) {
			head = (struct stuff *) malloc(sizeof(struct stuff));
			ptr = head;
			ptr->next = NULL;
		} else if (ptr == head) {
			head = (struct stuff *) malloc(sizeof(struct stuff));
			head->next = ptr;
			ptr = head;
		} else {
			tempptr = ptr;
			ptr = last;
			ptr->next = (struct stuff *)
				malloc(sizeof(struct stuff));
			ptr->next->next = tempptr;
			ptr = ptr->next;
		}
		strcpy(ptr->name, name);
		ptr->score = score;
	}
	num = 0;
	ptr = head;
	while (num++ < 10 && ptr != NULL) {
		sprintf(output, "%6d %-30s", ptr->score, ptr->name);
		writeln(s, output);
		ptr = ptr->next;
	}
	writeln(s, ":");	/* tell client that we are truly finished */
	ptr = head;
	while (ptr != NULL) {
		tempptr = ptr->next;
		free(ptr);
		ptr = tempptr;
	}
}


#if COMMENTED
/*
 * Figure out who wins.
 */

static char *score_arr[] = {
	"Nothing",
	"Pair",
	"Two Pair",
	"Three of a kind",
	"Straight",
	"Flush",
	"Full House",
	"Four of a kind",
	"Straight Flush",
	"Royal Flush"
};
#endif

/* all player information */
/* number of players */
/* show all cards regardless */
static int 
whowon(PLAYER * player, int n_players, int test)
{
	int     i;
	int     max = -1;	/* highest score */
	int     winner = 0;	/* which player won */
	char    temp[80], *p;	/* tell everybody what everybody has */
	void    tie();

	show_cards(player, n_players, TRUE, test); /*show all hands everywhere*/

	for (i = 0; i < n_players; i++)
		if (test || player[i].in) {
			p = evaluate_char(player, i);
			sprintf(temp, "R%d%s", i, p);
			tellall(player, n_players, temp);

			if (player[i].in && player[i].score > max) {
				max = player[i].score;
				winner = i;
			} else if (player[i].in && player[i].score == max) {
				tie(player, &winner, i);
			}
		}
	return winner;
}

/* all hands on deck */
/* may have to set new winner */
/* which player is challenging the winner */

static void 
tie(PLAYER * player, int *winner, int which)
{
	int     top1, top2;
	int     i;
	int     done = FALSE;
	int     oldwinner;

	oldwinner = *winner;
	while (!done) {
		top1 = top2 = -1;
		for (i = 12; i >= 0; i--) {
			if (top1 == -1 ||
			    player[which].count[i] > player[which].count[top1])
				top1 = i;
			if (top2 == -1 ||
			player[*winner].count[i] > player[*winner].count[top2])
				top2 = i;
		}
		if (top1 != top2) {
			if (top1 > top2)
				*winner = which;
			done = TRUE;
		} else if (player[which].count[top1] == 0) {	/* OH, NO! */
			printf("tie() failed to resolve - details follow\n");
			for (i = 0; i < 5; i++)
				printf("%c%c ", rank(player[which].cards[i]),
					color(player[which].cards[i]));
			for (i = 0; i < 5; i++)
				printf("%c%c ", rank(player[*winner].cards[i]),
					color(player[*winner].cards[i]));
			fflush(stdout);
			done = TRUE;
		} else {
			player[which].count[top1] = 0;
			player[*winner].count[top2] = 0;
		}
	}
	/* re-evaluate hands because counts have been messed ... */
	evaluate(player, oldwinner);
	evaluate(player, which);
}

int large, small;

void
find_multiplex(int *intp, int which, int neednext)
{
	int	i;

	if (neednext) { /* two pair case */
		for (i=0; i<13; i++) {
			if (intp[i] == 2) {
				small = i;
				break;
			}
		}
		for (i++; i<13; i++) {
			if (intp[i] == 2) {
				large = i;
				break;
			}
		}
		return;
	}

	for (i=0; i<13; i++) {
		if (intp[i] == which) {
			small = i;
			return;
		}
	}
}
	

/* which player's hand to evaluate */
char * 
evaluate_char(PLAYER *player, int which)
{
	static char evalbuf[64];
	int     kind[6];
	int     i, j;
	char	*cp;
	PLAYER	*pp;

	pp = &player[which];
	for (i = 0; i < 13; i++)
		pp->count[i] = 0;

	for (i = 0; i < 5; i++)
		pp->count[card(pp->cards[i])]++;
	for (i = 0; i < 6; i++)
		kind[i] = 0;
	for (i = 0; i < 13; i++)
		kind[pp->count[i]]++;

	cp = (char *) evalbuf;
	if (kind[4]) {
		pp->score = FOUR;
		find_multiplex(pp->count, 4, 0);
		sprintf(evalbuf, "%c Four of@a kind", rank(small));
	} else if (kind[3] && kind[2]) {
		pp->score = FULL;
		find_multiplex(pp->count, 3, 0);
		sprintf(evalbuf, "%c Full house", rank(small));
	} else if (kind[3]) {
		pp->score = THREE;
		find_multiplex(pp->count, 3, 0);
		sprintf(evalbuf, "%c Three of@a kind", rank(small));
	} else if (kind[2] == 2) {
		pp->score = TWOPAIR;
		find_multiplex(pp->count, 2, 1);
		sprintf(evalbuf, "%c%c Two Pair", rank(large), rank(small));
	} else if (kind[2]) {
		pp->score = PAIR;
		find_multiplex(pp->count, 2, 0);
		sprintf(evalbuf, "%c One Pair", rank(small));
	} else {
		pp->score = NOTHING;
		strcpy(evalbuf, "NOTHING");
	}

	if (pp->score == NOTHING) {
		for (i = 1; i < 5; i++)
			if (suit(pp->cards[i]) != suit(pp->cards[0]))
				break;

		if (i == 5) {
			pp->score = FLUSH;
			sprintf(evalbuf, "%s Flush",
				suits_names[suit(pp->cards[0])]);
		}

		i = -1;
		while (pp->count[++i] == 0)
			/* void */ ;
		j = i;
		while (++i < 13 && pp->count[i] == 1)
			/* void */ ;

		/* trash below is for A-5 straights */
		if (i - j == 5) {
			if (pp->score == FLUSH) {
				pp->score = STRAIGHT_FLUSH;
				sprintf(evalbuf, "Straight Flush");
			} else {
				pp->score = STRAIGHT;
				sprintf(evalbuf, "%c Straight",
					rank(pp->count[i-1]));
			}
		} else if ((i == 4) && (j == 0) && (pp->count[12] == 1)) {
			if (pp->score == FLUSH) {
				pp->score = STRAIGHT_FLUSH;
				sprintf(evalbuf, "Straight Flush");
			} else {
				pp->score = STRAIGHT;
				sprintf(evalbuf, "5 Straight");
			}
		}

		if (pp->score == STRAIGHT_FLUSH && i == 13) {
			pp->score = ROYAL_FLUSH;
			sprintf(evalbuf, "Royal flush");
		}
	}
	return evalbuf;
}

/* which player's hand to evaluate */
static void 
evaluate(PLAYER *player, int which)
{
	int     kind[6];
	int     i, j;
	PLAYER	*pp;

	pp = &player[which];
	for (i = 0; i < 13; i++)
		pp->count[i] = 0;

	for (i = 0; i < 5; i++)
		pp->count[card(pp->cards[i])]++;
	for (i = 0; i < 6; i++)
		kind[i] = 0;
	for (i = 0; i < 13; i++)
		kind[pp->count[i]]++;

	if (kind[4])
		pp->score = FOUR;
	else if (kind[3] && kind[2])
		pp->score = FULL;
	else if (kind[3])
		pp->score = THREE;
	else if (kind[2] == 2)
		pp->score = TWOPAIR;
	else if (kind[2])
		pp->score = PAIR;
	else
		pp->score = NOTHING;

	if (pp->score == NOTHING) {
		for (i = 1; i < 5; i++)
			if (suit(pp->cards[i]) != suit(pp->cards[0]))
				break;

		if (i == 5)
			pp->score = FLUSH;
		i = -1;
		while (pp->count[++i] == 0)
			/* void */ ;
		j = i;
		while (++i < 13 && pp->count[i] == 1)
			/* void */ ;

		/* trash below is for A-5 straights */
		if (i - j == 5 ||((i == 4) && (j == 0) && (pp->count[12] == 1)))
			if (pp->score == FLUSH)
				pp->score = STRAIGHT_FLUSH;
			else
				pp->score = STRAIGHT;

		if (pp->score == STRAIGHT_FLUSH && i == 13)
			pp->score = ROYAL_FLUSH;
	}
}
/*
 * Handle all the betting being done.
 */

/* all player information */
/* # of players */
/* current pot */
/* player who gets to bet first */

static int 
betting(PLAYER *player, int n_players, int *pot, int first)
{
	int     last_raiser = (-1);	/* player who made last raise */
	int     n_active = 0;	/* # players active in betting(haven't folded)*/
	int     n_raises = 0;	/* # of raises this hand so far */
	int     j = (-1);	/* player currently betting */
	int     i;
	int     total_bet;	/* total amount bet each player this hand */
	char    temp[80];
	int     done;		/* has player entered something valid yet? */
	int     amount;		/* how much more is player donating to pot? */
	int     cycle = FALSE;	/* has bet been around circle once yet? */
	int     x;
	int     rounds = 0;

	i = first;
	while (i != first || !cycle) {
		if (player[i].in) {
			if (j == -1) {
				total_bet = player[i].bet;
				j = i;
			}
			sprintf(temp, "B%d%d", i, player[i].bet);
			tellall(player, n_players, temp);
			sprintf(temp, "$%d%d", i, player[i].cash);
			tellall(player, n_players, temp);
			n_active++;
		}
		if (++i == n_players) {
			cycle = TRUE;
			i = 0;
		}
	}

	tellall(player, n_players, "U");
	while (n_active > 1 && j != last_raiser) {
		sprintf(temp, "T%d", j);
		tellall(player, n_players, temp);
		sprintf(temp, "P%d", *pot);
		tellall(player, n_players, temp);
		sprintf(temp, "B%d%d", j, player[j].bet);
		tellall(player, n_players, temp);
		tellall(player, n_players, "U");
		sprintf(temp, "$%d%d", j, player[j].cash);
		tellall(player, n_players, temp);

		if (j > 0) {	/* if not computer */
			sprintf(temp, "?%d", total_bet - player[j].bet);
			writeln(player[j].socket, temp);
		}
		tellall(player, n_players, "U");	/* update displays */

		/* get bet ... */

		done = FALSE;
		while (!done) {
			if (j > 0)
				readln(player[j].socket, temp);
			else
				computer_bet(player, n_players, n_active,
					n_raises, *pot, total_bet, temp);
			done = TRUE;

			switch (temp[0]) {
			case 'Q':
				player[j].wantsout = TRUE;
				done = FALSE;
				break;
			case 'S':
				player[j].sittingout = TRUE;
				done = FALSE;
				break;
			case 'F':
				player[j].in = FALSE;
				sprintf(temp, "H%d--fold--", j);
				tellall(player, n_players, temp);
				n_active--;
				break;
			case 'C':
				amount = total_bet - player[j].bet;
				*pot += amount;
				player[j].cash -= amount;
				player[j].bet = total_bet;
				sprintf(temp, "B%d%d", j, total_bet);
				tellall(player, n_players, temp);
				sprintf(temp, "H%dCall", j);
				tellall(player, n_players, temp);
				break;
			case 'R':
				if (++n_raises > n_players) {
					sprintf(temp,
					"MMax # of raises; Call or fold.");
					writeln(player[j].socket, temp);
					--j;
					break;
				}
				last_raiser = j;
				amount = (total_bet += (x = atoi(temp + 1)))
					- player[j].bet;

				*pot += amount;
				player[j].cash -= amount;
				player[j].bet = total_bet;
				sprintf(temp, "B%d%d", j, total_bet);
				tellall(player, n_players, temp);
				sprintf(temp, "H%dRaise %d", j, x);
				tellall(player, n_players, temp);
				break;
			case 'P':
				sprintf(temp, "H%dPass", j);
				tellall(player, n_players, temp);
				break;
			}
		}
		sprintf(temp, "$%d%d", j, player[j].cash);
		tellall(player, n_players, temp);

		sprintf(temp, "P%d", *pot);
		tellall(player, n_players, temp);
		tellall(player, n_players, "U");
		if (++j >= n_players) {
			++rounds;
			j = 0;
		}
		while (!player[j].in)
			if (++j >= n_players) {
				++rounds;
				j = 0;
			}
		if (last_raiser == -1
		    && (rounds > 1 || (rounds == 1 && j >= first)))
			break;	/* all passed or folded */
	}
	if (n_active <= 1)
		return j;
	else
		return -1;
}

/*
 * Basic card playing routines like shuffling, etc.
 */
#define	 MAXRAND 2147483648

/* return deck of shuffled cards here */
void 
shuffle_original(int *deck)
{
	int     temp[52];
	int     i, x;
	int     crandom(int limit);

	for (i = 0; i < 52; i++)
		temp[i] = i;
	for (i = 0; i < 52; i++) {
		while (temp[x = crandom(52)] == ERROR)
			/* void */ ;
		deck[i] = temp[x];
		temp[x] = ERROR;
	}
}

void
swap(int *a, int *b)
{
	int temp;

	temp = *a;
	*a = *b;
	*b = temp;
}

static void
shuffle(int *deck)
{
	int     i, j, swap_with;
	/*
	 * The pack is shuffled 500 times. The higher this value the more
	 * random is the pack, but it is slower between deals
	 */
	for (i = 0; i < 52; i++)
		deck[i] = i;

	for (i = 0; i < 500; i++) {
		for (j = 0; j < DECK_SIZE; j++) {
			swap_with = rand() % DECK_SIZE;
			swap(&deck[j], &deck[swap_with]);
		}
	}
}


int 
crandom(int limit)
{
	unsigned long random(void);
	unsigned long scale;
	int     x;

	scale = (unsigned long) MAXRAND / limit;
	x = random() / scale;
	if (x < 0)
		x = 0;		/* weird things do happen occasionally */
	return (x);
}

/*
 * Handle betting and drawing for the computer's hand.
 */

#define	min(x,y)	((x) > (y) ? (y) : (x))

static int bluffing = FALSE;	/* bluffing this hand? */
static int bluffmore = FALSE;	/* more bluff bets this hand? */

/* player information & hands */
/* number of players */
/* number of players still in current hand */
/* # raises so far this betting session */
/* size of pot */
/* current total bet */
/* return betting command here */

static void 
computer_bet(PLAYER * player, int n_players, int n_active,
	int n_raises, int pot, int total, char *temp)
{
	int     raise;	 /* amount to raise by */
	int     chance;	 /* rating of hand */
	int     top = (-1); /* highest card hand has most of */
	int     i;

	static int lastpot = 99999;/* trick to only get random once per hand */
	static int random_factor = 0;

	evaluate(player, 0);
	for (i = 12; i > 0; i--)
		if (top == -1 || player[0].count[i] > player[0].count[top])
			top = i;
	if (pot < lastpot) {	/* things to be done once per hand */
		random_factor = n_players * crandom(10);
		if (crandom(20) < 1)	/* bluff sometimes	 */
			bluffing = TRUE;	/* 1 out of 20 chance 	 */
		else
			bluffing = FALSE;
		bluffmore = bluffing;
	}
	lastpot = pot;
	chance = player[0].score * player[0].score * 10
		+ top + 1 + random_factor;
	if (bluffmore)
		chance += n_players * 50;
	if (((total < chance && crandom(10) < 6) || chance > total + 80)
	    && n_raises < n_players) {
		raise = min(RAISE_LIMIT,
			(player[0].score + crandom(3) - 1) * 10);
		if (raise <= 5)
			raise = 5;
		if (bluffing) {
			raise = RAISE_LIMIT - crandom(2) * 10;
			bluffmore = FALSE;
		}
		sprintf(temp, "R%d", min(RAISE_LIMIT, raise));
	} else if (total <= player[0].bet)
		sprintf(temp, "P");
	else if (player[0].score >= THREE /* 3 of a kind or better never folds*/
		 || (n_active == 2 && player[0].score > NOTHING)
		 || total < chance + 80 - 100 / n_players) {
		/* comparatively good hand */
		sprintf(temp, "C");
	} else
		sprintf(temp, "F");
}

static void 
computer_draw(PLAYER *player, char *temp)
{
	int     i, j;
	/* hand has already been evaluated by computer_bet() at least once. */

	strcpy(temp, "Dkkkkk");
	if (player[0].score < STRAIGHT)	/* higher hands require no discards */
		for (i = 0; i < 5; i++) {
			for (j = 0; j < 5; j++)
				if (i != j && card(player[0].cards[i])
				    == card(player[0].cards[j]))
					break;
			if (j == 5) {
				temp[i + 1] = 'd';	/* discard it */
			}
		}
	bluffmore = bluffing;
}
/*
 * Let each player who is still in draw as many as 5 cards to replace *
 * those he already has in his hand. This routines expects a command * of the
 * form Dxxxxx  from each player program, where x is either * 'k'  to keep
 * the card in that position, or * 'd'  to discard the card in that position.
 */


/* hands are in here */
/* # of players */
/* shuffled deck of cards */
/* next card to take from deck */
/* player who gets to draw first */

static void 
draw(PLAYER * player, int n_players, int *deck, int d, int first)
{
	int     i, c;
	char    temp[80];
	char    temp2[20];
	int     n;
	int     cycle = FALSE;

	i = first;
	while (i != first || !cycle) {
		if (player[i].in) {
			sprintf(temp, "T%d", i);
			tellall(player, n_players, temp);
			tellall(player, n_players, "U");

			if (i == 0)
				computer_draw(player, temp);
			else {
				writeln(player[i].socket, "D");
				readln(player[i].socket, temp);
			}
			if (temp[0] != 'D') {	/* what else could it be ?? */
				--i;	/* do this man again ! */
				continue;
			}
			n = 0;
			for (c = 0; c < 5; c++)
				if (temp[c + 1] == 'd') {
					player[i].cards[c] = deck[d++];
					n++;
				}

			if (n == 0)
				strcpy(temp2, "Draws none");
			else
				sprintf(temp2, "Draws %d", n);

			sprintf(temp, "H%d%s", i, temp2);
			tellall(player, n_players, temp);
			tellall(player, n_players, "U");

			if (i != 0) {
				show(player, temp, i, i);
				writeln(player[i].socket, "U");
			}
		}
		if (++i == n_players) {
			cycle = TRUE;
			i = 0;
		}
	}
}
/*
 * Tell each player everybody's name. Show each player only his own cards.
 */

/* all player info */
/* # active players */
/*
 * TRUE=show everybody's cards who is still in FALSE=show only player's cards
 * to himself
 */
/* show all cards whether in or out */
/* only used if 'all' is TRUE */

static void 
show_cards(PLAYER * player, int n_players, int all, int test)
{
	int     i, j, k;
	char    temp[10];
	void    show(PLAYER * player, char *temp, int i, int j);

	for (i = 1; i < n_players; i++)	/* don't show cards to computer */
		if (player[i].bet >= 0) {	/* if sitting out, be quiet */

			for (j = 0; j < n_players; j++) {
				/* tell player everyone's name */
				sprintf(temp, "N%d", j);
				write(player[i].socket, temp, 2);
				writeln(player[i].socket, player[j].name);
			}
			if (all) {
				for (k = 0; k < n_players; k++)
					if (test || player[k].in)
						show(player, temp, k, i);
			} else
				show(player, temp, i, i);

			writeln(player[i].socket, "U");/*update screen command*/
		}
}

/* whose cards to show */
/* who to show them to */
static void 
show(PLAYER *player, char *temp, int i, int j)
{
	int     c;

	sprintf(temp, "C%d", i);/* cards for player i */
	write(player[j].socket, temp, 2);
	for (c = 0; c < 5; c++) {
		/* hunsoo
		sprintf(temp, "%c%c ",
			rank(player[i].cards[c]), color(player[i].cards[c]));
		 */
		sprintf(temp, "%c%c ",
			card(player[i].cards[c]) + '0',
			suit(player[i].cards[c]) + '0');
		write(player[j].socket, temp, 3);
	}
	write(player[j].socket, "\n", 1);
}
